/*****************************************************************************
 *
 * Copyright (C) 2012  Florian Pose <fp@igh-essen.com>
 *
 * This file is part of the QtPdWidgets library.
 *
 * The QtPdWidgets library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * The QtPdWidgets library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the QtPdWidgets Library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "ScalarSubscriber.h"
using Pd::ScalarSubscriber;

#include <pdcom5/Exception.h>
#include <pdcom5/Subscriber.h>
#include <pdcom5/Subscription.h>

/****************************************************************************/

/* defined in header util.h */
PdCom::Transmission getTrans(double sampleTime)
{
    if (sampleTime > 0.0) {
        return { std::chrono::duration<double>(sampleTime) };
    }
    if (sampleTime == 0.0) {
        return { PdCom::event_mode };
    }
    return { PdCom::poll_mode };
}

/****************************************************************************/

struct Pd::ScalarSubscriber::Impl
{
    Impl(ScalarSubscriber *parent):
        parent(parent),
        filterConstant(0.0)
    {
    }

    ScalarSubscriber * const parent;

    struct ScalarSubscription;
    std::unique_ptr<ScalarSubscription> subscription;

    double filterConstant; /**< PT1 filter constant. */
};

/****************************************************************************/

struct Pd::ScalarSubscriber::Impl::ScalarSubscription:
    public PdCom::Subscriber, public PdCom::Subscription
{
    public:
        ScalarSubscription(ScalarSubscriber::Impl *parent, PdCom::Variable pv,
                const PdCom::Selector &selector, double sampleTime):
            Subscriber{getTrans(sampleTime)},
            Subscription{*this, pv, selector},
            parent{parent}
        {
        }

        ScalarSubscription(ScalarSubscriber::Impl *parent,
                PdCom::Process *process, const std::string &path,
                const PdCom::Selector &selector, double sampleTime):
            Subscriber{getTrans(sampleTime)},
            Subscription{*this, *process, path, selector},
            parent{parent}
        {
        }

    private:
        ScalarSubscriber::Impl *parent;

        void stateChanged(const PdCom::Subscription &) override
        {
#if DEBUG
            QString path;
            if (not getVariable().empty()) {
                path = getVariable().getPath().data();
            }
            qDebug() << this << __func__ << (int) getState() << path;
#endif

            if (getState() == PdCom::Subscription::State::Active) {
                // changed to active. If event mode, poll once.
                if (getTransmission() == PdCom::event_mode) {
                    poll(); // poll once to get initial value
                }
            }

            parent->parent->stateChange(getState());
        }

        void newValues(std::chrono::nanoseconds ts) override
        {
            parent->parent->newValues(ts);
        }
};

/****************************************************************************/

/** Constructor.
 */
ScalarSubscriber::ScalarSubscriber():
    scale{1.0},
    offset{0.0},
    impl{std::unique_ptr<Impl>{new Impl{this}}}
{
}

/****************************************************************************/

/** Destructor.
 */
ScalarSubscriber::~ScalarSubscriber()
{
    clearVariable();
}

/****************************************************************************/

/** Subscribes to a ProcessVariable.
 */
void ScalarSubscriber::setVariable(
        PdCom::Variable pv,
        const PdCom::Selector &selector,
        double sampleTime,
        double scale,
        double offset,
        double tau
        )
{
    clearVariable();

    if (pv.empty()) {
        return;
    }

    this->scale = scale;
    this->offset = offset;

    if (tau > 0.0 && sampleTime > 0.0) {
        impl->filterConstant = sampleTime / tau;
    } else {
        impl->filterConstant = 0.0;
    }

    try {
        impl->subscription = std::unique_ptr<Impl::ScalarSubscription>(
                new Impl::ScalarSubscription(impl.get(), pv, selector,
                    sampleTime));
    } catch (PdCom::Exception &e) {
        qCritical() << QString("Failed to subscribe to variable"
                " \"%1\" with sample time %2: %3")
            .arg(QString(pv.getPath().c_str()))
            .arg(sampleTime)
            .arg(e.what());
        return;
    }
}

/****************************************************************************/

/** Subscribes to a ProcessVariable.
 */
void ScalarSubscriber::setVariable(
        PdCom::Process *process,
        const QString &path,
        const PdCom::Selector &selector,
        double sampleTime,
        double scale,
        double offset,
        double tau
        )
{
    clearVariable();

    if (path.isEmpty() or not process) {
        return;
    }

    this->scale = scale;
    this->offset = offset;

    if (tau > 0.0 && sampleTime > 0.0) {
        impl->filterConstant = sampleTime / tau;
    } else {
        impl->filterConstant = 0.0;
    }

    try {
        impl->subscription = std::unique_ptr<Impl::ScalarSubscription>(
                new Impl::ScalarSubscription(impl.get(), process,
                    path.toLocal8Bit().constData(), selector, sampleTime));
    } catch (PdCom::Exception &e) {
        qCritical() << QString("Failed to subscribe to variable"
                " \"%1\" with sample time %2: %3")
            .arg(path)
            .arg(sampleTime)
            .arg(e.what());
        return;
    }
}

/****************************************************************************/

/** Unsubscribe from a Variable.
 */
void ScalarSubscriber::clearVariable()
{
    if (impl->subscription) {
        impl->subscription.reset();
        stateChange(PdCom::Subscription::State::Invalid);
    }
}

/****************************************************************************/

bool ScalarSubscriber::hasVariable() const
{
    return impl->subscription and
        not impl->subscription->getVariable().empty();
}

/****************************************************************************/

void ScalarSubscriber::stateChange(PdCom::Subscription::State)
{
}

/****************************************************************************/

double ScalarSubscriber::getFilterConstant() const
{
    return impl->filterConstant;
}

/****************************************************************************/

PdCom::Variable ScalarSubscriber::getVariable() const
{
    if (impl->subscription) {
        return impl->subscription->getVariable();
    }
    else {
        return PdCom::Variable();
    }
}

/****************************************************************************/

const void *ScalarSubscriber::ScalarSubscriber::getData() const
{
    if (impl->subscription) {
        return impl->subscription->getData();
    }
    else {
        return nullptr;
    }
}

/****************************************************************************/
